from fastapi import Query, HTTPException
from Bio.Restriction.Restriction_Dictionary import rest_dict


from ..dna_functions import (
    format_sequence_genbank,
    read_dsrecord_from_json,
)
from ..dna_utils import align_sanger_traces
from ..pydantic_models import (
    TextFileSequence,
    BaseCloningStrategy,
)
from ..get_router import get_router
from ..utils import api_version


router = get_router()


@router.get('/version')
async def get_version():
    return api_version()


@router.get('/restriction_enzyme_list', response_model=dict[str, list[str]])
async def get_restriction_enzyme_list():
    """Return the dictionary of restriction enzymes"""
    return {'enzyme_names': list(rest_dict.keys())}


@router.post(
    '/validate',
    summary='Validate a cloning strategy',
)
async def cloning_strategy_is_valid(
    cloning_strategy: BaseCloningStrategy,
) -> bool:
    """Validate a cloning strategy"""
    return True


@router.post('/rename_sequence', response_model=TextFileSequence)
async def rename_sequence(
    sequence: TextFileSequence,
    name: str = Query(..., description='The new name of the sequence.', pattern=r'^[^\s]+$'),
):
    """Rename a sequence"""
    dseqr = read_dsrecord_from_json(sequence)
    return format_sequence_genbank(dseqr, name)


@router.post('/align_sanger', response_model=list[str])
async def align_sanger(
    sequence: TextFileSequence,
    traces: list[str],
):
    """Align a list of sanger traces to a sequence"""

    dseqr = read_dsrecord_from_json(sequence)
    try:
        return align_sanger_traces(dseqr, traces)
    except RuntimeError as e:
        raise HTTPException(status_code=400, detail=str(e))



from itertools import product
from pydna.dseqrecord import Dseqrecord
from Bio.Data.IUPACData import ambiguous_dna_values
from Bio.Seq import reverse_complement
from .dna_utils import compute_regex_site, dseqrecord_finditer

# We create a dictionary to map ambiguous bases to their consensus base
# For example, ambigous_base_dict['ACGT'] -> 'N'
ambiguous_base_dict = {}
for ambiguous, bases in ambiguous_dna_values.items():
    ambiguous_base_dict[''.join(sorted(bases))] = ambiguous

# To handle N values
ambiguous_base_dict['N'] = 'N'

# This is the original loxP sequence, here for reference
LOXP_SEQUENCE = 'ATAACTTCGTATAGCATACATTATACGAAGTTAT'

loxP_sequences = [
    # https://blog.addgene.org/plasmids-101-cre-lox
    # loxP
    'ATAACTTCGTATANNNTANNNTATACGAAGTTAT',
    # PMID:12202778
    # lox66
    'ATAACTTCGTATANNNTANNNTATACGAACGGTA',
    # lox71
    'TACCGTTCGTATANNNTANNNTATACGAAGTTAT',
]

loxP_consensus = ''

for pos in range(len(LOXP_SEQUENCE)):
    all_letters = set(seq[pos] for seq in loxP_sequences)
    key = ''.join(sorted(all_letters))
    loxP_consensus += ambiguous_base_dict[key]

# We compute the regex for the forward and reverse loxP sequences
loxP_regex = (compute_regex_site(loxP_consensus), compute_regex_site(reverse_complement(loxP_consensus)))


def cre_loxP_overlap(x: Dseqrecord, y: Dseqrecord, _l: None = None) -> list[tuple[int, int, int]]:
    """Find matching loxP sites between two sequences."""
    out = list()
    for pattern in loxP_regex:
        matches_x = dseqrecord_finditer(pattern, x)
        matches_y = dseqrecord_finditer(pattern, y)

        for match_x, match_y in product(matches_x, matches_y):
            value_x = match_x.group()
            value_y = match_y.group()
            if value_x[13:21] == value_y[13:21]:
                out.append((match_x.start() + 13, match_y.start() + 13, 8))
    # Unique values (keeping the order)
    unique_out = []
    for item in out:
        if item not in unique_out:
            unique_out.append(item)
    return unique_out
